@@ -3,26 +3,29 @@
 NestedText: A Human Readable and Writable Data Format
 """
 
-# License {{{1
-# Copyright (c) 2020 Kenneth S. Kundert
+# MIT License {{{1
+# Copyright (c) 2020 Kenneth S. Kundert and Kale Kundert
 #
-# This program is free software: you can redistribute it and/or modify it under
-# the terms of the GNU General Public License as published by the Free Software
-# Foundation, either version 3 of the License, or (at your option) any later
-# version.
+# Permission is hereby granted, free of charge, to any person obtaining a copy
+# of this software and associated documentation files (the "Software"), to deal
+# in the Software without restriction, including without limitation the rights
+# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+# copies of the Software, and to permit persons to whom the Software is
+# furnished to do so, subject to the following conditions:
 #
-# This program is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
-# details.
+# The above copyright notice and this permission notice shall be included in all
+# copies or substantial portions of the Software.
 #
-# You should have received a copy of the GNU General Public License along with
-# this program.  If not, see http://www.gnu.org/licenses/.
+# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+# SOFTWARE.
 
 # Imports {{{1
-import re
 from inform import (
-    cull,
     full_stop,
     set_culprit,
     get_culprit,
@@ -31,19 +34,20 @@ from inform import (
     is_mapping,
     Error,
     Info,
-    InformantFactory,
 )
+import collections.abc
+import re
 
 
 # Globals {{{1
-__version__ = "0.5.0"
-__released__ = "2020-09-11"
-__all__ = ['loads', 'dumps', 'NestedTextError']
+__version__ = "0.6.0"
+__released__ = "2020-09-26"
+__all__ = ['load', 'loads', 'dump', 'dumps', 'NestedTextError']
 
 
 # Exception {{{1
 class NestedTextError(Error, ValueError):
-    '''
+    r'''
     :func:`loads()` and :func:`dumps()` both raise
     *NestedTextError* when they discover an error. *NestedTextError* subclasses both
     the Python *ValueError* and the *Error* exception from *Inform*.
@@ -58,13 +62,9 @@ class NestedTextError(Error, ValueError):
         The source of the *NestedText* content, if given. This is often a
         filename.
 
-    doc:
-
-        The *NestedText* content passed to :func:`loads`.
-
     line:
 
-        The line of *NestedText* content where the problem was found.
+        The text of the line of *NestedText* content where the problem was found.
 
     lineno:
 
@@ -74,6 +74,10 @@ class NestedTextError(Error, ValueError):
 
         The number of the character where the problem was found on *line*.
 
+    prev_line:
+
+        The text of the line immediately before where the problem was found.
+
     template:
 
         The possibly parameterized text used for the error message.
@@ -84,7 +88,7 @@ class NestedTextError(Error, ValueError):
     .. code-block:: python
 
         >>> from textwrap import dedent
-        >>> import nestedtext
+        >>> import nestedtext as nt
 
         >>> content = dedent("""
         ...     name1: value1
@@ -93,8 +97,8 @@ class NestedTextError(Error, ValueError):
         ... """).strip()
 
         >>> try:
-        ...     print(nestedtext.loads(content))
-        ... except nestedtext.NestedTextError as e:
+        ...     print(nt.loads(content))
+        ... except nt.NestedTextError as e:
         ...     print(str(e))
         2: duplicate key: name1.
 
@@ -103,22 +107,47 @@ class NestedTextError(Error, ValueError):
     *inform*'s conventions::
 
         >> try:
-        ..     print(nestedtext.loads(content))
-        .. except nestedtext.NestedTextError as e:
+        ..     print(nt.loads(content))
+        .. except nt.NestedTextError as e:
         ..     e.report()
         error: 2: duplicate key: name1.
             «name1: value2»
-             ↑
+             ▲
 
     The *terminate* method prints the message directly and exits::
 
         >> try:
-        ..     print(nestedtext.loads(content))
-        .. except nestedtext.NestedTextError as e:
+        ..     print(nt.loads(content))
+        .. except nt.NestedTextError as e:
         ..     e.terminate()
         error: 2: duplicate key: name1.
             «name1: value2»
-             ↑
+             ▲
+
+    With exceptions generated from :func:`load` or :func:`loads` you may see
+    extra lines at the end of the message that show the problematic lines if
+    you have the exception report itself as above.  Those extra lines are
+    referred to as the codicil and they can be very helpful in illustrating the
+    actual problem. You do not get them if you simply cast the exception to a
+    string, but you can access them using :meth:`NestedTextError.get_codicil`.
+    The codicil or codicils are returned as a tuple.  You should join them with
+    newlines before printing them.
+
+    .. code-block:: python
+
+        >>> try:
+        ...     print(nt.loads(content))
+        ... except nt.NestedTextError as e:
+        ...     print(e.get_message())
+        ...     print(*e.get_codicil(), sep="\n")
+        duplicate key: name1.
+           1 «name1: value1»
+           2 «name1: value2»
+              ▲
+
+    Note the « and » characters in the codicil. They delimit the extend of the
+    text on each line and help you see troublesome leading or trailing white
+    space.
 
     Exceptions produced by *NestedText* contain a *template* attribute that
     contains the basic text of the message. You can change this message by
@@ -130,115 +159,33 @@ class NestedTextError(Error, ValueError):
     .. code-block:: python
 
         >>> try:
-        ...     print(nestedtext.loads(content))
-        ... except nestedtext.NestedTextError as e:
+        ...     print(nt.loads(content))
+        ... except nt.NestedTextError as e:
         ...     template = None
         ...     if e.template == 'duplicate key: {}.':
         ...         template = 'llave duplicada: {}.'
         ...     print(e.render(template=template))
         2: llave duplicada: name1.
 
-    When you have the exception report itself, you see up to two extra lines in
-    the message that are used to display the line and the location where the
-    problem was found.  Those extra lines are referred to as the codicil. You
-    do not get them if you simply cast the exception to a string, but you can
-    access them using :meth:`NestedTextError.get_codicil`.  There is
-    an additional method,
-    :meth:`NestedTextError.get_extended_codicil` that also shows the
-    source of the problem, but with extra context::
-
-        >> try:
-        ..     print(nestedtext.loads(content))
-        .. except nestedtext.NestedTextError as e:
-        ..     e.report(codicil=e.get_extended_codicil())
-        error: 2: duplicate key: name1.
-            1> name1: value1
-            2> name1: value2
-               ↑
-            2> name3: value3
-
-    Both the normal and the extended codicils are returned as tuples.
-    You should join them with newlines before printing them.
     '''
 
-    def get_extended_codicil(self, codicil=None):
-        """Get the extended codicil.
-
-        The extended codicil is like the normal codicil in that it shows the
-        line where the problem was found and points to the suspect character.
-        However, it also shows a few lines of surrounding context.
-
-        This method counts on the fact that no user specified codicils have been
-        added to the exception; that the only codicil available is the one added
-        by *NestedText* itself.  Any additional codicils can be add by passing
-        them into this method.
-
-        Args:
-            codicil (string or tuple of strings):
-                An extra codicil or collection of codicils.  Appended to the
-                return value without modifying the cached codicil.
-
-        Returns:
-            The codicil argument is appended to the exception's codicil and the
-            combination is returned. The return value is always in the form of a
-            tuple even if there is only one component.
-        """
-        exception_codicil = self.kwargs.get('codicil', ())
-        if not is_collection(exception_codicil):
-            exception_codicil = (exception_codicil,)  # pragma: no cover
-
-        # Like the normal codicil, but provides a few lines of surrounding
-        # context.
-        try:
-            lineno = self.lineno
-            doc = self.doc
-            colno = self.colno
-            lines_before = doc.split('\n')[max(lineno-2, 0):lineno]
-            lines = []
-            for i, l in zip(range(lineno-len(lines_before), lineno), lines_before):
-                lines.append(f'{i+1:>4} «{l}»')
-            lines_before = '\n'.join(lines)
-            lines_after = doc.split('\n')[lineno:lineno+1]
-            lines = []
-            for i, l in zip(range(lineno, lineno + len(lines_after)), lines_after):
-                lines.append(f'{i+1:>4} «{l}»')
-            lines_after = '\n'.join(lines)
-            exception_codicil = '\n'.join(cull([
-                lines_before,
-                '      ' + (colno*' ') + '↑',
-                lines_after
-            ]))
-            exception_codicil = (exception_codicil, )
-        except Exception:
-            exception_codicil = self.get_codicil()
-
-        if codicil:
-            if not is_collection(codicil):
-                codicil = (codicil,)
-            return exception_codicil + codicil
-        return exception_codicil
-
 
 # NestedText Reader {{{1
 # Converts NestedText into Python data hierarchies.
 
 # constants {{{2
-dict_tag = ": "
-quoted = "|".join([
+# build regex use to split dict items
+colon_space = ": "
+colon_newline = ":\n"
+splitters = [
     r'"[^"\n]*"',  # "string"
     r"'[^'\n]*'",  # 'string'
-])
-splitters = (
-    quoted,           # "string" or 'string' (must be first)
-    dict_tag,         # key/value separator in dictionary item
-)
+    colon_space,
+    colon_newline,
+]
 splitter = re.compile("(" + "|".join(f"(?:{s})" for s in splitters) + ")")
 
 
-# debugging utilities {{{2
-highlight = InformantFactory(message_color='blue')
-
-
 # report {{{2
 def report(message, line, *args, colno=None, **kwargs):
     message = full_stop(message)
@@ -248,13 +195,22 @@ def report(message, line, *args, colno=None, **kwargs):
     if line:
         kwargs['culprit'] = get_culprit(line.lineno)
         if colno is not None:
-            kwargs['codicil'] = f"«{line.text}»\n {colno*' '}↑"
+            # build codicil that shows both the line and the preceding line
+            if line.prev_line is not None:
+                codicil = [f'{line.lineno-1:>4} «{line.prev_line}»']
+            else:
+                codicil = []
+            codicil += [
+                f'{line.lineno:>4} «{line.text}»',
+                '      ' + (colno*' ') + '▲',
+            ]
+            kwargs['codicil'] = '\n'.join(codicil)
             kwargs['colno'] = colno
         else:
-            kwargs['codicil'] = f"«{line.text}»"
+            kwargs['codicil'] = f'{line.lineno:>4} «{line.text}»'
         kwargs['line'] = line.text
         kwargs['lineno'] = line.lineno
-        kwargs['doc'] = line.content
+        kwargs['prev_line'] = line.prev_line
     else:
         kwargs['culprit'] = culprits  # pragma: no cover
     raise NestedTextError(template=message, *args, **kwargs)
@@ -281,12 +237,9 @@ def dequote(s):
 
 # Lines class {{{2
 class Lines:
-    class Line(Info):
-        pass
-
     # constructor {{{3
-    def __init__(self, content):
-        self.content = content
+    def __init__(self, lines):
+        self.lines = lines
         self.generator = self.read_lines()
         self.next_line = True
         while self.next_line:
@@ -294,39 +247,59 @@ class Lines:
             if self.next_line and self.next_line.kind not in ["blank", "comment"]:
                 return
 
+    # Line class {{{3
+    class Line(Info):
+        pass
+
     # read_lines() {{{3
     def read_lines(self):
-        for lineno, line in enumerate(self.content.splitlines()):
+        prev_line = None
+        for lineno, line in enumerate(self.lines):
             depth = None
             key = None
             value = None
+
+            # last line may not have trailing newline, which we count on
+            if line[-1:] != '\n':
+                line += '\n'
+
+            # compute indentation
             stripped = line.lstrip()
             depth = len(line) - len(stripped)
+
+            # determine line type and extract values
             if stripped == "":
                 kind = "blank"
-                value = "\n"
+                value = None
+                depth = None
             elif stripped[:1] == "#":
                 kind = "comment"
-                value = line[1:].strip()
-            elif stripped == '-' or stripped.startswith('- '):
+                value = line[1:-1].strip()
+                depth = None
+            elif stripped == '-\n' or stripped.startswith('- '):
                 kind = "list item"
-                value = stripped[2:]
-            elif stripped == '>' or stripped.startswith('> '):
+                value = stripped[2:-1]
+            elif stripped == '>\n' or stripped.startswith('> '):
                 kind = "string"
-                value = line[depth+2:]
+                value = line[depth+2:-1]
             else:
-                # add space so tags with empty values can be specified w/o space
-                components = splitter.split(line + " ")
-                if dict_tag in components:
-                    kind = "dict item"
-                    split_loc = components.index(dict_tag)
+                components = splitter.split(line)
+                kind = "dict item"
+                if colon_space in components:
+                    split_loc = components.index(colon_space)
                     key = dequote("".join(components[:split_loc]))
                     value = "".join(components[split_loc + 1:])
-                    value = value[:-1]  # remove space added above
+                elif colon_newline in components:
+                    # :\n is always second to last component and followed by ''
+                    key = dequote("".join(components[:-2]))
+                    value = "\n"
                 else:
                     kind = "unrecognized"
                     value = line
+                value = value[:-1]
 
+            # bundle information about line
+            line = line[:-1]  # remove trailing newline
             the_line = self.Line(
                 text = line,
                 lineno = lineno+1,
@@ -334,8 +307,9 @@ class Lines:
                 depth = depth,
                 key = key,
                 value = value,
-                content = self.content
+                prev_line = prev_line,
             )
+            prev_line = line
 
             # check the indent for non-spaces
             if depth:
@@ -462,15 +436,27 @@ def read_string(lines, depth):
     return "\n".join(data)
 
 
+# read_all() {{{2
+def read_all(lines, source, on_dup):
+    if callable(on_dup):
+        on_dup = dict(_callback_func=on_dup)
+
+    with set_culprit(source):
+        lines = Lines(lines)
+
+        if lines.type_of_next():
+            return read_value(lines, 0, on_dup)
+
+
 # loads() {{{2
-def loads(content, source=None, on_dup=None):
-    '''
-    Loads NestedText from string.
+def loads(content, source=None, *, on_dup=None):
+    r'''
+    Loads *NestedText* from string.
 
     Args:
         content (str):
             String that contains encoded data.
-        source (str):
+        source (str or Path):
             If given, this string is attached to any error messages as the
             culprit. It is otherwise unused. Is often the name of the file that
             originally contained the NestedText content.
@@ -492,13 +478,16 @@ def loads(content, source=None, on_dup=None):
     Returns:
         The extracted data.  If content is empty, None is returned.
 
+    Raises:
+        NestedTextError: if there is a problem in the *NextedText* content.
+
     Examples:
 
         *NestedText* is specified to *loads* in the form of a string:
 
         .. code-block:: python
 
-            >>> import nestedtext
+            >>> import nestedtext as nt
 
             >>> contents = """
             ... name: Kristel Templeton
@@ -507,8 +496,8 @@ def loads(content, source=None, on_dup=None):
             ... """
 
             >>> try:
-            ...     data = nestedtext.loads(contents)
-            ... except nestedtext.NestedTextError as e:
+            ...     data = nt.loads(contents)
+            ... except nt.NestedTextError as e:
             ...     e.terminate()
 
             >>> print(data)
@@ -525,15 +514,14 @@ def loads(content, source=None, on_dup=None):
             >>> filename = 'examples/duplicate-keys.nt'
             >>> try:
             ...     with open(filename, encoding='utf-8') as f:
-            ...         addresses = nestedtext.loads(f.read(), filename)
-            ... except nestedtext.NestedTextError as e:
-            ...     print(str(e))
-            ...     print(e.get_extended_codicil()[0])
+            ...         addresses = nt.loads(f.read(), filename)
+            ... except nt.NestedTextError as e:
+            ...     print(e.render())
+            ...     print(*e.get_codicil(), sep="\n")
             examples/duplicate-keys.nt, 5: duplicate key: name.
                4 «name:»
                5 «name:»
-                  ↑
-               6 «»
+                  ▲
 
         Notice in the above example the encoding is explicitly specified as
         'utf-8'.  *NestedText* files should always be read and written using
@@ -552,15 +540,15 @@ def loads(content, source=None, on_dup=None):
             ... name: value 5
             ... """
 
-            >>> print(nestedtext.loads(content))
+            >>> print(nt.loads(content))
             Traceback (most recent call last):
             ...
             nestedtext.NestedTextError: 3: duplicate key: key.
 
-            >>> print(nestedtext.loads(content, on_dup='ignore'))
+            >>> print(nt.loads(content, on_dup='ignore'))
             {'key': 'value 1', 'name': 'value 4'}
 
-            >>> print(nestedtext.loads(content, on_dup='replace'))
+            >>> print(nt.loads(content, on_dup='replace'))
             {'key': 'value 3', 'name': 'value 5'}
 
             >>> def de_dup(key, value, data, state):
@@ -569,23 +557,125 @@ def loads(content, source=None, on_dup=None):
             ...     state[key] += 1
             ...     return f"{key}#{state[key]}"
 
-            >>> print(nestedtext.loads(content, on_dup=de_dup))
+            >>> print(nt.loads(content, on_dup=de_dup))
             {'key': 'value 1', 'key#2': 'value 2', 'key#3': 'value 3', 'name': 'value 4', 'name#2': 'value 5'}
 
     '''
-    if callable(on_dup):
-        on_dup = dict(_callback_func=on_dup)
 
-    with set_culprit(source):
-        lines = Lines(content)
+    return read_all(content.splitlines(True), source, on_dup)
 
-        if lines.type_of_next():
-            return read_value(lines, 0, on_dup)
+
+# load() {{{2
+def load(f=None, on_dup=None):
+    r'''
+    Loads *NestedText* from file or stream.
+
+    Is the same as :func:`loads` except the *NextedText* is accessed by reading
+    a file rather than directly from a string. It does not keep the full
+    contents of the file in memory and so is more memory efficient with large
+    files.
+
+    Args:
+        f (str, os.PathLike, io.TextIOBase, collections.abc.Iterator):
+            The file to read the *NestedText* content from.  This can be
+            specified either as a path (e.g. a string or a `pathlib.Path`),
+            as a text IO object (e.g. an open file), or as an iterator.  If a
+            path is given, the file will be opened, read, and closed.  If an IO
+            object is given, it will be read and not closed; utf-8 encoding
+            should be used..  If an iterator is given, it should generate full
+            lines in the same manner that iterating on a file descriptor would.
+        on_dup:
+            See :func:`loads` description of this argument.
+
+    Returns:
+        The extracted data.  If content is empty, None is returned.
+
+    Raises:
+        NestedTextError: if there is a problem in the *NextedText* content.
+        OSError: if there is a problem opening the file.
+
+    Examples:
+
+        Load from a path specified as a string:
+
+        .. code-block:: python
+
+            >>> import nestedtext as nt
+            >>> print(open('examples/groceries.nt').read())
+            - Bread
+            - Peanut butter
+            - Jam
+            <BLANKLINE>
+
+            >>> nt.load('examples/groceries.nt')
+            ['Bread', 'Peanut butter', 'Jam']
+
+        Load from a `pathlib.Path`:
+
+        .. code-block:: python
+
+            >>> from pathlib import Path
+            >>> nt.load(Path('examples/groceries.nt'))
+            ['Bread', 'Peanut butter', 'Jam']
+
+        Load from an open file object:
+
+        .. code-block:: python
+
+            >>> with open('examples/groceries.nt') as f:
+            ...     nt.load(f)
+            ...
+            ['Bread', 'Peanut butter', 'Jam']
+
+    '''
+
+    # Do not invoke the read method as that would read in the entire contents of
+    # the file, possibly consuming a lot of memory. Instead pass the file
+    # pointer into read_all(), it will iterate through the lines, discarding
+    # them once they are no longer needed, which reduces the memory usage.
+
+    if isinstance(f, collections.abc.Iterator):
+        source = getattr(f, 'name', None)
+        return read_all(f, source, on_dup)
+    else:
+        source = str(f)
+        with open(f, encoding='utf-8') as fp:
+            return read_all(fp, source, on_dup)
 
 
 # NestedText Writer {{{1
 # Converts Python data hierarchies to NestedText.
 
+
+# render_key {{{2
+def render_key(s):
+    if not is_str(s):
+        raise NestedTextError(template='keys must be strings.', culprit=s)
+    stripped = s.strip(' ')
+    if '\n' in s:
+        raise NestedTextError(
+            s,
+            template='keys must not contain newlines.',
+            culprit=repr(s)
+        )
+    if (
+        len(stripped) < len(s)
+        or s[:1] == "#"
+        or s.startswith("- ")
+        or s.startswith("> ")
+        or ': ' in s
+        or s[:1] + s[-1:] in ['""', "''"]
+    ):
+        if '"' in s and "'" in s:
+            raise NestedTextError(
+                s,
+                template = """keys that require quoting must not contain both " and '.""",
+                culprit = s,
+            )
+        return repr(s)
+    return s
+
+
 # add_leader {{{2
 def add_leader(s, leader):
     # split into separate lines
@@ -598,13 +688,23 @@ def add_leader(s, leader):
     )
 
 
+# add_prefix {{{2
+def add_prefix(prefix, suffix):
+    # A simple formatting of dict and list items will result in a space
+    # after the colon or dash if the value is placed on next line.
+    # This, function simply eliminates that space.
+    if not suffix or suffix.startswith("\n"):
+        return prefix + suffix
+    return prefix + " " + suffix
+
+
 # dumps {{{2
 def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level=0):
-    """Recursively convert object to string with reasonable formatting.
+    """Recursively convert object to *NestedText* string.
 
     Args:
         obj:
-            The object to convert.
+            The object to convert to *NestedText*.
         sort_keys (bool or func):
             Dictionary items are sorted by their key if *sort_keys* is true.
             If a function is passed in, it is used as the key function.
@@ -630,6 +730,12 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             not specified by the user, but can be useful in unusual situations
             to specify an initial indent.
 
+    Returns:
+        The *NestedText* content.
+
+    Raises:
+        NestedTextError: if there is a problem in the input data.
+
     Examples:
         This example writes to a string, but it is common to write to a file.
         The file name and extension are arbitrary. However, by convention a
@@ -637,7 +743,7 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
 
         .. code-block:: python
 
-            >>> import nestedtext
+            >>> import nestedtext as nt
 
             >>> data = {
             ...     'name': 'Kristel Templeton',
@@ -646,8 +752,8 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             ... }
 
             >>> try:
-            ...     print(nestedtext.dumps(data))
-            ... except nestedtext.NestedTextError as e:
+            ...     print(nt.dumps(data))
+            ... except nt.NestedTextError as e:
             ...     print(str(e))
             name: Kristel Templeton
             sex: female
@@ -668,16 +774,16 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             >>> data = {'key': 42, 'value': 3.1415926, 'valid': True}
 
             >>> try:
-            ...     print(nestedtext.dumps(data))
-            ... except nestedtext.NestedTextError as e:
+            ...     print(nt.dumps(data))
+            ... except nt.NestedTextError as e:
             ...     print(str(e))
             key: 42
             value: 3.1415926
             valid: True
 
             >>> try:
-            ...     print(nestedtext.dumps(data, default='strict'))
-            ... except nestedtext.NestedTextError as e:
+            ...     print(nt.dumps(data, default='strict'))
+            ... except nt.NestedTextError as e:
             ...     print(str(e))
             42: unsupported type.
 
@@ -696,13 +802,13 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             ...         return self.color
 
             >>> data['house'] = Color('red')
-            >>> print(nestedtext.dumps(data, default=repr))
+            >>> print(nt.dumps(data, default=repr))
             key: 42
             value: 3.1415926
             valid: True
             house: Color('red')
 
-            >>> print(nestedtext.dumps(data, default=str))
+            >>> print(nt.dumps(data, default=str))
             key: 42
             value: 3.1415926
             valid: True
@@ -721,8 +827,8 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             ... }
 
             >>> try:
-            ...    print(nestedtext.dumps(data, renderers=renderers))
-            ... except nestedtext.NestedTextError as e:
+            ...    print(nt.dumps(data, renderers=renderers))
+            ... except nt.NestedTextError as e:
             ...     print(str(e))
             key: 0x2a
             value: 3.14
@@ -742,8 +848,8 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             ... }
 
             >>> try:
-            ...    print(nestedtext.dumps(data, renderers=renderers))
-            ... except nestedtext.NestedTextError as e:
+            ...    print(nt.dumps(data, renderers=renderers))
+            ... except nt.NestedTextError as e:
             ...     print(str(e))
             3.1415926: unsupported type.
 
@@ -785,40 +891,6 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             level = level + 1
         )
 
-    # render string
-    def render_key(s):
-        stripped = s.strip(' ')
-        if '\n' in s:
-            raise NestedTextError(
-                s,
-                template='keys must not contain newlines.',
-                culprit=repr(s)
-            )
-        if (
-            len(stripped) < len(s)
-            or s[:1] == "#"
-            or s.startswith("- ")
-            or s.startswith("> ")
-            or ': ' in s
-            or s[:1] + s[-1:] in ['""', "''"]
-        ):
-            if '"' in s and "'" in s:
-                raise NestedTextError(
-                    s,
-                    template = """keys that require quoting must not contain both " and '.""",
-                    culprit = s,
-                )
-            return repr(s)
-        return s
-
-    def add_prefix(prefix, suffix):
-        # A simple formatting of dict and list items will result in a space
-        # after the colon or dash if the value is placed on next line.
-        # This, function simply eliminates that space.
-        if not suffix or suffix.startswith("\n"):
-            return prefix + suffix
-        return prefix + " " + suffix
-
     # render content
     assert indent > 0
     error = None
@@ -847,6 +919,8 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             need_indented_block = True
         else:
             content = obj
+            if level == 0:
+                content = add_leader(content, '> ').strip()
     elif is_a_scalar(obj):
         content = str(obj)
     elif default and callable(default):
@@ -861,3 +935,85 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
         raise NestedTextError(obj, template=error, culprit=repr(obj))
 
     return content
+
+
+# dump {{{2
+def dump(obj, f, **kwargs):
+    """Write the *NestedText* representation of the given object to the given file.
+
+    Args:
+        obj:
+            The object to convert to *NestedText*.
+        f (str, os.PathLike, io.TextIOBase):
+            The file to write the *NestedText* content to.  The file can be
+            specified either as a path (e.g. a string or a `pathlib.Path`) or
+            as a text IO instance (e.g. an open file).  If a path is given, the
+            will be opened, written, and closed.  If an IO object is given, it
+            must have been opened in a mode that allows writing (e.g.
+            ``open(path, 'w')``), if applicable.  It will be written and not
+            closed.
+
+            The name used for the file is arbitrary but it is tradition to use a
+            .nt suffix.
+
+        kwargs:
+            See :func:`dumps` for optional arguments.
+
+    Returns:
+        The *NestedText* content.
+
+    Raises:
+        NestedTextError: if there is a problem in the input data.
+        OSError: if there is a problem opening the file.
+
+    Examples:
+
+        This example writes to a pointer to an open file.
+
+        .. code-block:: python
+
+            >>> import nestedtext as nt
+            >>> from inform import fatal, os_error
+
+            >>> data = {
+            ...     'name': 'Kristel Templeton',
+            ...     'sex': 'female',
+            ...     'age': '74',
+            ... }
+
+            >>> try:
+            ...     with open('data.nt', 'w', encoding='utf-8') as f:
+            ...         nt.dump(data, f)
+            ... except nt.NestedTextError as e:
+            ...     fatal(e)
+            ... except OSError as e:
+            ...     fatal(os_error(e))
+
+        This example writes to a file specified by file name.
+
+        .. code-block:: python
+
+            >>> try:
+            ...     nt.dump(data, 'data.nt')
+            ... except nt.NestedTextError as e:
+            ...     fatal(e)
+            ... except OSError as e:
+            ...     fatal(os_error(e))
+            >>> data = {'key': 42, 'value': 3.1415926, 'valid': True}
+
+    """
+    content = dumps(obj, **kwargs)
+
+    # Avoid nested try-except blocks, since they lead to chained exceptions
+    # (e.g. if the file isn't found, etc.) that unnecessarily complicate the
+    # stack trace.
+
+    try:
+        f.write(content)
+    except AttributeError:
+        pass
+    else:
+        return
+
+    with open(f, 'w', encoding='utf-8') as fp:
+        fp.write(content)
